import torch
import math
class STCLossFunction(torch.autograd.Function):
    @staticmethod
    def create_stc_graph(target, blank_idx, lam):
        g = gtn.Graph(False)
        L, S  = len(target), 2 * len(target) + 1
        for l in range(S):
            idx = (l - 1) // 2
            g.add_node(l == 0, l == S - 1 or l == S - 2)
            label = target[idx] if l % 2 else blank_idx
            if label == blank_idx:
                g.add_arc(l, l, label)
            if l > 0:
                g.add_arc(l - 1, l, label)
            if l % 2 and l > 1 and label != target[idx - 1]:
                g.add_arc(l - 2, l, label)
        for l in range(L+1):
            p1, p2 = 2 * l - 1, 2 * l
            star_idx = 80   
            
            c1 = g.add_node(False, l == L);
            idx = star_idx if l == L else (star_idx + target[l])
            if p1 >=0:
                g.add_arc(p1, c1, idx, idx, math.log(lam))
            g.add_arc(p2, c1, idx, idx, math.log(lam))
            g.add_arc(c1, c1, idx, idx, math.log(lam))
            if l < L:
                g.add_arc(c1, 2 * l + 1, target[l]);
            g.add_arc(c1, p2, blank_idx); 
        g.arc_sort(False)
        return g


    @staticmethod
    def forward(ctx, log_probs, targets, lam, blank_idx=0, reduction="none"):
        B, T, C = log_probs.shape
        losses, scales, emissions_graphs = [None] * B,  [None] * B, [None] * B

        def process(b):
            # create emission graph
            g_emissions = gtn.linear_graph(T, C, gtn.Device(gtn.CPU), log_probs.requires_grad)
            cpu_data = log_probs[b].cpu().contiguous()
            g_emissions.set_weights(cpu_data.data_ptr())

            # create criterion graph
            g_criterion = STCLossFunction.create_stc_graph(targets[b], blank_idx, lam)
            # compose the graphs
            g_loss = gtn.negate(
                gtn.forward_score(gtn.compose(g_criterion, g_emissions))
            )

            scale = 1.0
            if reduction == "mean":
                scale = 1.0 / T if T > 0 else scale
            elif reduction != "none":
                raise ValueError("invalid value for reduction '" + str(reduction) + "'")

            # Save for backward:
            losses[b] = g_loss
            scales[b] = scale
            emissions_graphs[b] = g_emissions

        gtn.parallel_for(process, range(B))

        ctx.auxiliary_data = (losses, scales, emissions_graphs, log_probs.shape)
        loss = torch.tensor([losses[b].item() * scales[b] for b in range(B)])
        return torch.mean(loss.cuda() if log_probs.is_cuda else loss)

    @staticmethod
    def backward(ctx, grad_output):
        losses, scales, emissions_graphs, in_shape = ctx.auxiliary_data
        B, T, C = in_shape
        input_grad = torch.empty((B, T, C))

        def process(b):
            gtn.backward(losses[b], False)
            emissions = emissions_graphs[b]
            grad = emissions.grad().weights_to_numpy()
            input_grad[b] = torch.from_numpy(grad).view(1, T, C) * scales[b]

        gtn.parallel_for(process, range(B))

        if grad_output.is_cuda:
            input_grad = input_grad.cuda()
        input_grad *= grad_output / B

        return (
            input_grad,
            None,  # targets
            None,  # lam
            None,  # blank_idx
            None,  # reduction
        )
STCLoss = STCLossFunction.apply